"""
This file contains learning rate schedulers for tensorflow optimization.
"""

###############################################################################
# initial imports and set-up:

import numpy as np

# tensorflow imports:
import tensorflow as tf
from tensorflow.python.keras.callbacks import Callback
from tensorflow.python.platform import tf_logging as logging

###############################################################################
# Exponential decay:


class ExponentialDecayAnnealer():

    def __init__(self, start, end, roll_off_step, steps):
        self.start = start
        self.end = end
        self.roll_off_step = float(roll_off_step)
        self.steps = float(steps)
        self.n = 0

    def step(self):
        self.n += 1
        return self.start / (
            1. + (self.end / (self.start - self.end))**((self.n - self.roll_off_step) /
                                                        (self.roll_off_step - self.steps)))


class ExponentialDecayScheduler(Callback):

    def __init__(self, lr_max, lr_min, roll_off_step, steps):
        """
        Exponentially decaying learning rate.
        
        :param lr_max: maximum learning rate
        :param lr_min: minimum earning rate
        :param roll_off_step: step at which the scheduler starts rolling off
        :param steps: total number of steps
        """
        
        super(ExponentialDecayScheduler, self).__init__()

        self.step = 0
        self.Annealer = ExponentialDecayAnnealer(lr_max, lr_min, roll_off_step, steps)
        self.lrs = []

    def on_train_begin(self, logs=None):
        """ """
        self.step = 0
        self.set_lr(self.Annealer.start)

    def on_train_batch_begin(self, batch, logs=None):
        """ """
        self.lrs.append(self.get_lr())

    def on_train_batch_end(self, batch, logs=None):
        """ """
        self.step += 1
        self.set_lr(self.Annealer.step())

    def get_lr(self):
        """ """
        try:
            return tf.keras.backend.get_value(self.model.optimizer.lr)
        except AttributeError:
            return None

    def set_lr(self, lr):
        """ """
        try:
            tf.keras.backend.set_value(self.model.optimizer.lr, lr)
        except AttributeError:
            pass  # ignore


###############################################################################
# Power law decay:


class PowerLawDecayAnnealer():

    def __init__(self, start, end, power, steps):
        self.start = start
        self.end = end
        self.p = power
        self.steps = float(steps)
        self.n = 0

    def step(self):
        self.n += 1
        return self.start / (self.n / (self.steps * (self.end / self.start)**(1 / self.p)))**self.p


class PowerLawDecayScheduler(Callback):

    def __init__(self, lr_max, lr_min, power, steps):
        """
        Power law decaying learning rate.
            
        :param lr_max: maximum learning rate
        :param lr_min: minimum earning rate
        :param power: power law index
        :param steps: total number of steps
        """

        super(PowerLawDecayScheduler, self).__init__()

        self.step = 0
        self.Annealer = PowerLawDecayAnnealer(lr_max, lr_min, power, steps)
        self.lrs = []

    def on_train_begin(self, logs=None):
        """ """
        self.step = 0
        self.set_lr(self.Annealer.start)

    def on_train_batch_begin(self, batch, logs=None):
        """ """
        self.lrs.append(self.get_lr())

    def on_train_batch_end(self, batch, logs=None):
        """ """
        self.step += 1
        self.set_lr(self.Annealer.step())

    def get_lr(self):
        """ """
        try:
            return tf.keras.backend.get_value(self.model.optimizer.lr)
        except AttributeError:
            return None

    def set_lr(self, lr):
        """ """
        try:
            tf.keras.backend.set_value(self.model.optimizer.lr, lr)
        except AttributeError:
            pass  # ignore

###############################################################################
# Step decay:


class StepDecayAnnealer():

    def __init__(self, start=None, change_every=None, steps=None, steps_per_epoch=None, boundaries=None, values=None):
        self.start = start
        self.steps_per_epoch = steps_per_epoch
        if boundaries is not None:
            self.boundaries = boundaries
            self.values = values
        else:
            self.ch = change_every
            total_changes = int(steps / self.ch)
            self.end = self.start / (10**total_changes)
            self.steps = float(steps)

            self.boundaries = [self.ch * i for i in range(1, total_changes)]
            self.values = [self.start / (10**i) for i in range(0, total_changes)]

        self.n = 0

    def step(self):
        self.n += 1
        for i in range(len(self.boundaries)):
            if self.n < self.boundaries[i] * self.steps_per_epoch:
                return self.values[i]
            else:
                pass
        return self.values[-1]


class StepDecayScheduler(Callback):

    def __init__(self, lr_max=None, change_every=None, steps=None, steps_per_epoch=None, boundaries=None, values=None):
        """
        Learning rate step function changes.
        
        :param lr_max:
        :param change_every:
        :param steps:
        :param steps_per_epoch:
        :param boundaries:
        :param values:
        """

        super(StepDecayScheduler, self).__init__()

        self.step = 0
        self.Annealer = StepDecayAnnealer(lr_max, change_every, steps, steps_per_epoch, boundaries, values)
        self.lrs = []

    def on_train_begin(self, logs=None):
        """ """
        self.step = 0
        self.set_lr(self.Annealer.start)

    def on_train_batch_begin(self, batch, logs=None):
        """ """
        self.lrs.append(self.get_lr())

    def on_train_batch_end(self, batch, logs=None):
        """ """
        self.step += 1
        self.set_lr(self.Annealer.step())

    def get_lr(self):
        """ """
        try:
            return tf.keras.backend.get_value(self.model.optimizer.lr)
        except AttributeError:
            return None

    def set_lr(self, lr):
        """ """
        try:
            tf.keras.backend.set_value(self.model.optimizer.lr, lr)
        except AttributeError:
            pass  # ignore


###############################################################################
# Adaptive learning rate (loss slope) and early stopping:


class LRAdaptLossSlopeEarlyStop(Callback):

    def __init__(
            self,
            monitor="val_loss",
            factor=1. / np.sqrt(10.),
            patience=25,
            cooldown=10,
            verbose=0,
            min_lr=1e-5,
            threshold=0.0,
            **kwargs,
    ):
        """
        Adaptive reduction of learning rate when likelihood improvement stalls for a given number of epochs.

        :param monitor:
        :param factor:
        :param patience:
        :param cooldown:
        :param verbose:
        :param min_lr:
        """

        super().__init__()

        self.monitor = monitor
        if factor >= 1.0:
            raise ValueError("LRAdaptLossSlopeEarlyStop does not support a factor >= 1.0. Got {factor}")

        self.factor = factor
        self.min_lr = min_lr
        self.patience = patience
        self.verbose = verbose
        self.cooldown = cooldown
        self.threshold = threshold

        self._reset()

    def _reset(self):
        self.cooldown_counter = 0
        self.wait = 0
        self.last_losses = []

    def on_train_begin(self, logs=None):
        """ """
        self._reset()

    def on_epoch_end(self, epoch, logs=None):
        """ """
        logs = logs or {}
        logs["lr"] = tf.keras.backend.get_value(self.model.optimizer.lr)
        current = logs.get(self.monitor)
        if current is None:
            logging.warning(
                "Learning rate reduction is conditioned on metric `%s` "
                "which is not available. Available metrics are: %s",
                self.monitor,
                ",".join(list(logs.keys())),
            )
        else:
            if self.cooldown_counter > 0:
                self.cooldown_counter -= 1
                self.wait = 0
            else:
                self.last_losses.append(current)
                self.wait += 1
                if self.wait >= self.patience:
                    a = np.polyfit(np.arange(self.patience), self.last_losses[-self.patience:], 1)[0] # fits a line to `val_loss` in the last `patience` epochs
                    if a > self.threshold: # tests if val_loss is going up
                        old_lr = tf.keras.backend.get_value(self.model.optimizer.lr)
                        if old_lr > np.float32(self.min_lr):
                            new_lr = old_lr * self.factor
                            new_lr = max(new_lr, self.min_lr)
                            tf.keras.backend.set_value(self.model.optimizer.lr, new_lr)
                            if self.verbose > 0:
                                tf.print(
                                    f"\nEpoch {epoch +1}: "
                                    "LRAdaptLossSlopeEarlyStop reducing "
                                    f"learning rate to {new_lr}.")
                            self.cooldown_counter = self.cooldown
                            self.wait = 0
                            self.last_losses = []
                        else:
                            self.model.stop_training = True
                            


"""
This file contains the definition of the trainable bijectors that are needed to define normalizing flows.
"""

###############################################################################
# initial imports and set-up:

import numpy as np
import pickle
from collections.abc import Iterable
import functools
import collections

# tensorflow imports:
import tensorflow as tf
import tensorflow_probability as tfp
from tensorflow_probability.python.bijectors import bijector as bijector_lib
from tensorflow.keras.layers import Input, Lambda, Dense
from tensorflow.keras.models import Model, Sequential

# tensorflow internal imports:
from tensorflow_probability.python.internal import parameter_properties
from tensorflow_probability.python.internal import dtype_util
from tensorflow_probability.python.internal import tensor_util
from tensorflow_probability.python.internal import tensorshape_util

# tensorflow aliases:
tfb = tfp.bijectors
tfd = tfp.distributions

# local imports:
from .. import utilities as utils
from . import fixed_bijectors as fixed_bijectors

###############################################################################
# utility function to generate random permutations with minimum stack variance:


def min_var_permutations(d, n, min_number=10000):
    """
    Find a random permutation sequence that has the minimum (sample) variance between components.
    This is useful for MAFs, in which we want to concatenate several triangular transformations.

    :param d: dimension of the problem
    :param n: number of dimensions
    :param min_number: number of random trials
    :return: the permutation that has less variance
    """
    permutation = None
    identity = np.arange(d)
    perm_var = np.inf
    for _ in range(max(min_number, 2 * d * n)):
        # draw the permutation ensemble:
        _n_perm = 0
        _temp_perm = []
        while _n_perm < n:
            _temp = np.random.permutation(d)
            if not np.all(_temp == identity):
                _temp_perm.append(_temp)
                _n_perm += 1
        # calculate variance:
        _temp_var = np.var(np.sum(_temp_perm, axis=0))
        # save minimum:
        if _temp_var < perm_var:
            perm_var = _temp_var
            permutation = _temp_perm
    #
    return permutation


###############################################################################
# generic class:


class TrainableTransformation(object):

    def save(self, path):
        """
        Save to file(s) the state of the bijector.
        """
        raise NotImplementedError

    @classmethod
    def load(cls, path, **kwargs):
        """
        Load the trainable bijector.
        """
        raise NotImplementedError


###############################################################################
# class to build a scaling, rotation and shift bijector:

class ScaleRotoShift(tfb.Bijector):

    def __init__(
            self,
            dimension,
            scale=True,
            roto=True,
            shift=True,
            validate_args=False,
            initializer='glorot_uniform',
            name='Affine',
            dtype=tf.float32):
        """
        Bijector performing a shift, scaling and rotation.
        Note that scale is exponential so that we can do unconstrained optimization.
        The bijector is initialized to identity but can be changed with optional argument.
        This bijector does not use the Cholesky decomposition since we need guarantee of
        strictly positive definiteness and invertibility.

        :param dimension: (int) number of dimensions.
        :param scale: (bool) include (or not) scaling.
        :param roto: (bool) include (or not) rotations.
        :param shift: (bool) include (or not) shifts.
        :param validate_args: validate input arguments or not.
        :param initializer: initializer for the bijector, defaults to zeros (identity).
        :param name: name of the bijector.
        :param dtype: data type for the bijector, defaults to tf.float32.      
        :reference: https://arxiv.org/abs/1906.00587
        """

        parameters = dict(locals())

        with tf.name_scope(name) as name:

            self.dimension = dimension
            if shift:
                self._shift = tfp.layers.VariableLayer(
                    dimension, initializer=initializer, dtype=dtype, name=name + '_shift')
            else:
                self._shift = lambda _: tf.zeros(dimension, dtype=dtype, name=name + '_shift')
            if scale:
                self._scalevec = tfp.layers.VariableLayer(
                    dimension, initializer=initializer, dtype=dtype, name=name + '_scale')
            else:
                self._scalevec = lambda _: tf.zeros(dimension, dtype=dtype, name=name + '_scale')
            if roto:
                self._rotvec = tfp.layers.VariableLayer(
                    dimension * (dimension - 1) // 2,
                    initializer=initializer,
                    trainable=True,
                    dtype=dtype,
                    name=name + '_roto')
            else:
                self._rotvec = lambda _: tf.zeros(dimension * (dimension - 1) // 2, dtype=dtype, name=name + '_roto')

            super(ScaleRotoShift, self).__init__(
                forward_min_event_ndims=0,
                is_constant_jacobian=False,
                validate_args=validate_args,
                parameters=parameters,
                dtype=dtype,
                name=name)

    @property
    def shift(self):
        return self._shift

    @classmethod
    def _is_increasing(cls):
        return True

    def _getaff_invaff(self, x):
        L = tf.zeros((self.dimension, self.dimension), dtype=self.dtype)
        L = tf.tensor_scatter_nd_update(L, np.array(np.tril_indices(self.dimension, -1)).T, self._rotvec(x))
        Lambda2 = tf.linalg.diag(tf.math.exp(self._scalevec(x)))
        val_update = tf.ones(self.dimension)
        L = tf.tensor_scatter_nd_update(L, np.array(np.diag_indices(self.dimension)).T, val_update)
        Q, R = tf.linalg.qr(L)
        Phi2 = tf.linalg.matmul(L, tf.linalg.inv(R))
        self.aff = tf.linalg.matmul(tf.linalg.matmul(tf.transpose(Phi2), Lambda2), Phi2)
        self.invaff = tf.linalg.inv(self.aff)
        valdet = tf.linalg.logdet(self.aff)
        self.logdet = valdet / self.dimension

    def _forward(self, x):
        self._getaff_invaff(x)
        _aff = self.aff
        return tf.transpose(tf.linalg.matmul(_aff, tf.transpose(x))) + self._shift(x)[None, :]

    def _inverse(self, y):
        self._getaff_invaff(y)
        _invaff = self.invaff
        return tf.transpose(tf.linalg.matmul(_invaff, tf.transpose(y - self._shift(y)[None, :])))

    def _forward_log_det_jacobian(self, x):
        self._getaff_invaff(x)
        return self.logdet

    def _inverse_log_det_jacobian(self, y):
        return -self._forward_log_det_jacobian(self._inverse(y))

    @classmethod
    def _parameter_properties(cls, dtype):
        return {'shift': parameter_properties.ParameterProperties()}

###############################################################################
# class to build a circular spline:

_SplineShared = collections.namedtuple(
    'SplineShared', 'range_min,range_max,out_of_bounds,out_of_bounds_up,out_of_bounds_dn,x_k,y_k,d_k,d_kp1,h_k,w_k,s_k')

class CircularRationalQuadraticSpline(tfb.RationalQuadraticSpline):
    """
    Rational quadratic spline that has non-zero slope at the boundaries.
    """
    
    def __init__(self,
                 bin_widths,
                 bin_heights,
                 knot_slopes,
                 boundary_knot_slope,
                 range_min=-1,
                 range_max=1,
                 validate_args=False,
                 name=None):
        """
        """

        with tf.name_scope(name or 'RationalQuadraticSpline') as name:

            dtype = dtype_util.common_dtype(
                    [bin_widths, bin_heights, knot_slopes, range_min, boundary_knot_slope],
                    dtype_hint=tf.float32)
            
            self._boundary_knot_slope = tensor_util.convert_nonref_to_tensor(
                boundary_knot_slope, dtype=dtype, name='boundary_knot_slope')
            
            self._range_max = tensor_util.convert_nonref_to_tensor(
                range_max, dtype=dtype, name='range_max')

            super(CircularRationalQuadraticSpline, self).__init__(
                bin_widths=bin_widths,
                bin_heights=bin_heights,
                knot_slopes=knot_slopes,
                range_min=range_min,
                validate_args=validate_args,
                name=name)

    @property
    def range_max(self):
        return self._range_max

    def _compute_shared(self, x=None, y=None):
        """
        See documentation of tfb.RationalQuadraticSpline._compute_shared.
        """

        assert (x is None) != (y is None)
        is_x = x is not None


        range_min = tf.convert_to_tensor(self.range_min, name='range_min')
        range_max = tf.convert_to_tensor(self.range_max, name='range_max')  

        kx = tfb.rational_quadratic_spline._knot_positions(self.bin_widths, range_min)
        ky = tfb.rational_quadratic_spline._knot_positions(self.bin_heights, range_min)
        kd = tf.concat([tf.expand_dims(self._boundary_knot_slope, -1), 
                        self.knot_slopes, 
                        tf.expand_dims(self._boundary_knot_slope, -1)], axis=-1)

        kx_or_ky = kx if is_x else ky
                
        kx_or_ky_min = kx_or_ky[..., 0]
        kx_or_ky_max = kx_or_ky[..., -1]
        
        x_or_y = x if is_x else y
        
        out_of_bounds_up = x_or_y >= kx_or_ky_max
        out_of_bounds_dn = x_or_y <= kx_or_ky_min
        out_of_bounds = out_of_bounds_dn | out_of_bounds_up

        x_or_y = tf.where(out_of_bounds, kx_or_ky_min, x_or_y)

        shape = functools.reduce(
            tf.broadcast_dynamic_shape,
            (
                tf.shape(x_or_y[..., tf.newaxis]),  # Add a n_knots dim.
                tf.shape(kx),
                tf.shape(ky),
                tf.shape(kd)))

        bc_x_or_y = tf.broadcast_to(x_or_y, shape[:-1])
        bc_kx = tf.broadcast_to(kx, shape)
        bc_ky = tf.broadcast_to(ky, shape)
        bc_kd = tf.broadcast_to(kd, shape)
        bc_kx_or_ky = bc_kx if is_x else bc_ky
        indices = tf.maximum(
            tf.zeros([], dtype=tf.int64),
            tf.searchsorted(
                bc_kx_or_ky[..., :-1],
                bc_x_or_y[..., tf.newaxis],
                side='right',
                out_type=tf.int64) - 1)

        def gather_squeeze(params, indices):
            rank = tensorshape_util.rank(indices.shape)
            if rank is None:
                raise ValueError('`indices` must have statically known rank.')
            return tf.gather(params, indices, axis=-1, batch_dims=rank - 1)[..., 0]

        x_k = gather_squeeze(bc_kx, indices)
        x_kp1 = gather_squeeze(bc_kx, indices + 1)
        y_k = gather_squeeze(bc_ky, indices)
        y_kp1 = gather_squeeze(bc_ky, indices + 1)
        d_k = gather_squeeze(bc_kd, indices)
        d_kp1 = gather_squeeze(bc_kd, indices + 1)
        h_k = y_kp1 - y_k
        w_k = x_kp1 - x_k
        s_k = h_k / w_k
                        
        return _SplineShared(
            range_min=range_min,
            range_max=range_max,
            out_of_bounds=out_of_bounds,
            out_of_bounds_up=out_of_bounds_up,
            out_of_bounds_dn=out_of_bounds_dn,
            x_k=x_k,
            y_k=y_k,
            d_k=d_k,
            d_kp1=d_kp1,
            h_k=h_k,
            w_k=w_k,
            s_k=s_k)
        
    def _forward(self, x):
        """Compute the forward transformation (Appendix A.1)."""
        d = self._compute_shared(x=x)
        relx = (x - d.x_k) / d.w_k
        spline_val = (
            d.y_k + ((d.h_k * (d.s_k * relx**2 + d.d_k * relx * (1 - relx))) /
                    (d.s_k + (d.d_kp1 + d.d_k - 2 * d.s_k) * relx * (1 - relx))))
        # apply bounds:
        y_val = tf.where(d.out_of_bounds_up, 
                         self._boundary_knot_slope * (x-d.range_max) +d.range_max, 
                         spline_val)
        y_val = tf.where(d.out_of_bounds_dn, 
                         self._boundary_knot_slope * (x-d.range_min) +d.range_min, 
                         y_val)
        return y_val

    def _inverse(self, y):
        """Compute the inverse transformation (Appendix A.3)."""
        d = self._compute_shared(y=y)
        rely = tf.where(d.out_of_bounds, tf.zeros([], dtype=y.dtype), y - d.y_k)
        term2 = rely * (d.d_kp1 + d.d_k - 2 * d.s_k)
        # These terms are the a, b, c terms of the quadratic formula.
        a = d.h_k * (d.s_k - d.d_k) + term2
        b = d.h_k * d.d_k - term2
        c = -d.s_k * rely
        # The expression used here has better numerical behavior for small 4*a*c.
        relx = tf.where(
            tf.equal(rely, 0), tf.zeros([], dtype=a.dtype),
            (2 * c) / (-b - tf.sqrt(b**2 - 4 * a * c)))

        # apply bounds:
        x_val = tf.where(d.out_of_bounds_up, 
                         self._boundary_knot_slope * (y-d.range_max) +d.range_max, 
                         relx * d.w_k + d.x_k)
        x_val = tf.where(d.out_of_bounds_dn, 
                         self._boundary_knot_slope * (y-d.range_min) +d.range_min, 
                         x_val) 
        return x_val

    def _forward_log_det_jacobian(self, x):
        """Compute the forward derivative (Appendix A.2)."""
        d = self._compute_shared(x=x)
        relx = (x - d.x_k) / d.w_k
        relx = tf.where(d.out_of_bounds, tf.constant(.5, x.dtype), relx)
       
        grad = (
            2 * tf.math.log(d.s_k) +
            tf.math.log(d.d_kp1 * relx**2 + 2 * d.s_k * relx * (1 - relx) +  # newln
                        d.d_k * (1 - relx)**2) -
            2 * tf.math.log((d.d_kp1 + d.d_k - 2 * d.s_k) * relx *
                            (1 - relx) + d.s_k))
        return tf.where(d.out_of_bounds, tf.math.log(self._boundary_knot_slope), grad)

###############################################################################
# helper class to build a spline-autoregressive flow, base spline class:


class SplineHelper(tfb.MaskedAutoregressiveFlow):

    def __init__(
        self,
        shift_and_log_scale_fn=None,
        bijector_fn=None,
        is_constant_jacobian=False,
        validate_args=False,
        unroll_loop=False,
        event_ndims=1,
        name=None,
        spline_knots=8,
        range_max=5.,
        range_min=None,
        equispaced_x_knots=False,
        equispaced_y_knots=False,
        slope_min=0.0001,
        min_bin_width=0.0,
        min_bin_height=0.0,
        slope_std=None,
        softplus_alpha=10.,
        dtype=tf.float32,
        ):
        """
        """
        parameters = dict(locals())
        name = name or 'spline_flow'

        # set ranges:
        if range_min is None:
            assert range_max > 0.
            range_min = -range_max
        interval_width = range_max - range_min
        
        # equispaced knots handling:
        if equispaced_x_knots or equispaced_y_knots:
            delta = (range_max-range_min)/(spline_knots)
        if equispaced_x_knots and equispaced_y_knots:
            raise ValueError('Cannot have both x and y knots equispaced.')

        with tf.name_scope(name) as name:
            self._unroll_loop = unroll_loop
            self._event_ndims = event_ndims
            if bool(shift_and_log_scale_fn) == bool(bijector_fn):
                raise ValueError('Exactly one of `shift_and_log_scale_fn` and '
                                 '`bijector_fn` should be specified.')
            if shift_and_log_scale_fn:

                def _bijector_fn(x, **condition_kwargs):

                    def reshape(params):
                        
                        factor = tf.cast(interval_width, dtype=dtype)

                        # get x grid:
                        if not equispaced_x_knots:
                            bin_widths = min_bin_width +params[..., :spline_knots]
                            bin_widths = tf.math.softmax(bin_widths)
                            bin_widths = tf.math.scalar_mul(factor, bin_widths)
                        else:
                            bin_widths = delta * tf.ones_like(params[..., :spline_knots])

                        # get y grid:
                        if not equispaced_y_knots:
                            if equispaced_x_knots:    
                                bin_heights = min_bin_height +params[..., :spline_knots]
                            else:                                              
                                bin_heights = min_bin_height +params[..., spline_knots:spline_knots * 2]
                            bin_heights = tf.math.softmax(bin_heights)
                            bin_heights = tf.math.scalar_mul(factor, bin_heights)
                        else:
                            bin_heights = delta * tf.ones_like(params[..., :spline_knots])
                                
                        # get knot slopes:
                        _start_idx = 2*spline_knots
                        if equispaced_x_knots:
                            _start_idx = _start_idx - spline_knots
                        if equispaced_y_knots:
                            _start_idx = _start_idx - spline_knots
                        knot_slopes = params[..., _start_idx:]
                        
                        # treat different cases:
                        if slope_std is None:
                            # sigmoid:
                            knot_slopes = 2. * tf.math.sigmoid(knot_slopes)
                            # enforce minimum slope:
                            knot_slopes = slope_min + tf.math.scalar_mul(2.-slope_min,knot_slopes)
                        else:                        
                            # deviations around finite differences
                            avg_slope = (bin_heights[...,1:]+bin_heights[...,:-1])/(bin_widths[...,1:]+bin_widths[...,:-1]) # finite diff
                            knot_slopes = avg_slope + tf.math.scalar_mul(slope_std, tf.math.tanh(knot_slopes)) # small deviations around finite diff
                            knot_slopes = tf.math.softplus(knot_slopes*softplus_alpha)/softplus_alpha # ensure slope is positive

                        return bin_widths, bin_heights, knot_slopes

                    params = shift_and_log_scale_fn(x, **condition_kwargs)
                    bin_widths, bin_heights, knot_slopes = reshape(params)

                    return tfb.RationalQuadraticSpline(
                        bin_widths=bin_widths,
                        bin_heights=bin_heights,
                        knot_slopes=knot_slopes,
                        range_min=range_min,
                        validate_args=False)

                bijector_fn = _bijector_fn

            # Still do this assignment for variable tracking.
            self._shift_and_log_scale_fn = shift_and_log_scale_fn
            self._bijector_fn = bijector_fn

            # Call the init method of the Bijector class and not that of MaskedAutoregressiveFlow which we are overriding
            bijector_lib.Bijector.__init__(
                self,
                forward_min_event_ndims=self._event_ndims,
                is_constant_jacobian=is_constant_jacobian,
                validate_args=validate_args,
                parameters=parameters,
                name=name)


class CircularSplineHelper(tfb.MaskedAutoregressiveFlow):

    def __init__(
        self,
        shift_and_log_scale_fn=None,
        bijector_fn=None,
        is_constant_jacobian=False,
        validate_args=False,
        unroll_loop=False,
        event_ndims=1,
        name=None,
        spline_knots=8,
        range_max=5.,
        range_min=None,
        equispaced_x_knots=False,
        equispaced_y_knots=False,
        slope_min=0.0001,
        min_bin_width=0.0,
        min_bin_height=0.0,
        slope_std=None,
        softplus_alpha=10.,
        dtype=tf.float32,
        ):
        """
        """
        parameters = dict(locals())
        name = name or 'circular_spline_flow'

        # set ranges:
        if range_min is None:
            assert range_max > 0.
            range_min = -range_max
        interval_width = range_max - range_min
        
        # equispaced knots handling:
        if equispaced_x_knots or equispaced_y_knots:
            delta = (range_max-range_min)/(spline_knots)
        if equispaced_x_knots and equispaced_y_knots:
            raise ValueError('Cannot have both x and y knots equispaced.')

        with tf.name_scope(name) as name:
            self._unroll_loop = unroll_loop
            self._event_ndims = event_ndims
            if bool(shift_and_log_scale_fn) == bool(bijector_fn):
                raise ValueError('Exactly one of `shift_and_log_scale_fn` and '
                                 '`bijector_fn` should be specified.')
            if shift_and_log_scale_fn:

                def _bijector_fn(x, **condition_kwargs):

                    def reshape(params):
                        
                        factor = tf.cast(interval_width, dtype=dtype)

                        # get x grid:
                        if not equispaced_x_knots:
                            bin_widths = min_bin_width +params[..., :spline_knots]
                            bin_widths = tf.math.softmax(bin_widths)
                            bin_widths = tf.math.scalar_mul(factor, bin_widths)
                        else:
                            bin_widths = delta * tf.ones_like(params[..., :spline_knots])

                        # get y grid:
                        if not equispaced_y_knots:
                            if equispaced_x_knots:    
                                bin_heights = min_bin_height +params[..., :spline_knots]
                            else:                                              
                                bin_heights = min_bin_height +params[..., spline_knots:spline_knots * 2]
                            bin_heights = tf.math.softmax(bin_heights)
                            bin_heights = tf.math.scalar_mul(factor, bin_heights)
                        else:
                            bin_heights = delta * tf.ones_like(params[..., :spline_knots])

                        # get knot slopes:
                        _start_idx = 2*spline_knots
                        if equispaced_x_knots:
                            _start_idx -= spline_knots
                        if equispaced_y_knots:
                            _start_idx -= spline_knots
                        knot_slopes = params[..., _start_idx:]

                        # treat different cases:
                        if slope_std is None:
                            # sigmoid:
                            knot_slopes = 2. * tf.math.sigmoid(knot_slopes)
                            # enforce minimum slope:
                            knot_slopes = slope_min + tf.math.scalar_mul(2.-slope_min,knot_slopes)
                            # get boundary slope:
                            boundary_knot_slope = knot_slopes[..., -1]                        
                            knot_slopes = knot_slopes[..., :-1]
                        else:
                            raise NotImplementedError

                        return bin_widths, bin_heights, knot_slopes, boundary_knot_slope

                    params = shift_and_log_scale_fn(x, **condition_kwargs)
                    bin_widths, bin_heights, knot_slopes, boundary_knot_slope = reshape(params)
                                                                                
                    temp_bijector = CircularRationalQuadraticSpline(
                                        bin_widths=bin_widths,
                                        bin_heights=bin_heights,
                                        knot_slopes=knot_slopes,
                                        boundary_knot_slope=boundary_knot_slope,
                                        range_min=range_min,
                                        range_max=range_max,
                                        validate_args=False)

                    return temp_bijector

                bijector_fn = _bijector_fn

            # Still do this assignment for variable tracking.
            self._shift_and_log_scale_fn = shift_and_log_scale_fn
            self._bijector_fn = bijector_fn

            # Call the init method of the Bijector class and not that of MaskedAutoregressiveFlow which we are overriding
            bijector_lib.Bijector.__init__(
                self,
                forward_min_event_ndims=self._event_ndims,
                is_constant_jacobian=is_constant_jacobian,
                validate_args=validate_args,
                parameters=parameters,
                name=name)
            

###############################################################################
# Make separate NNs for each dimension:

def build_nn(dim_in, dim_out, hidden_units, activation=tf.math.asinh, **kwargs):
    if len(hidden_units) == 0 or dim_in == 0:
        model = Sequential(Dense(dim_out, activation=None, input_shape=(dim_in,)))
    else:
        model = Sequential()
        model.add(Dense(hidden_units[0], activation=activation, input_shape=(dim_in,), **kwargs))
        for n in hidden_units[1:]:
            model.add(Dense(n, activation=activation, **kwargs))
        model.add(Dense(dim_out, activation=None, **kwargs))
    return model

@tf.function
def const_zeros(tensor, dim):
    batch_size = tf.shape(tensor)[0]
    constant = tf.zeros(dim)
    constant = tf.expand_dims(constant, axis=0)
    return tf.broadcast_to(constant, shape=(batch_size, dim))

def build_AR_model(num_params, transf_params, hidden_units=[], scale_with_dim=True, identity_dims=None, **kwargs):
    x = Input(num_params)
    params = []
    for dim in range(num_params):
        if identity_dims is not None and dim in identity_dims:
            # params.append(Lambda(lambda _x: tf.broadcast_to(tf.zeros(dim), (tf.shape(_x)[0], dim)))(x))
            params.append(Lambda(lambda _x: const_zeros(_x, transf_params))(x))
        else:
            if dim==0:
                _h = []
            else:
                if scale_with_dim:
                    _h = [int(np.ceil(h * (dim+1) / num_params)) for h in hidden_units]
                else:
                    _h = hidden_units
            nn = build_nn(dim, transf_params, hidden_units=_h, **kwargs)
            params.append(nn(x[..., :dim]))
    params = Lambda(lambda x: tf.stack(x, axis=-2))(params)
    return Model(x, params)


###############################################################################
# helper class to build a spline-autoregressive flow:


class AutoregressiveFlow(TrainableTransformation):
    """
    """

    def __init__(
            self,
            num_params,
            transformation_type='affine',  # 'affine' or 'spline'
            autoregressive_type='masked',  # 'masked' or 'flex'
            n_transformations=None,
            hidden_units=None,
            periodic_params=None,
            activation=tf.math.asinh,
            kernel_initializer=None,
            permutations=True,
            scale_roto_shift=False,
            parameters_min=None,
            parameters_max=None,
            # spline parameters:
            map_to_unitcube=False,
            spline_knots=8,
            range_max=5.,
            equispaced_x_knots=False,
            equispaced_y_knots=False,
            # other parameters:
            autoregressive_scale_with_dim=True,
            autoregressive_identity_dims=None,
            int_np_prec=np.int32,
            np_prec=np.float32,
            feedback=0,
            **kwargs):
        """
        :param num_params: number of parameters of the distribution.
        :param transformation_type: type of transformation, either 'affine' or 'spline'.
        :param autoregressive_type: type of autoregressive network, either 'masked' or 'flex'.
        :param n_transformations: number of transformations to concatenate.
        :param hidden_units: list of hidden units for the autoregressive network.
        :param periodic_params: bool list of parameters that are periodic.
        """

        if n_transformations is None:
            n_transformations = int(np.ceil(2 * np.log2(num_params)) + 2)
        event_shape = (num_params,)

        if hidden_units is None:
            hidden_units = [num_params * 2] * 2

        if isinstance(transformation_type, str):
            _transformation_types = [transformation_type] * n_transformations
            _autoregressive_types = [autoregressive_type] * n_transformations
        else:
            _transformation_types = transformation_type
            _autoregressive_types = autoregressive_type
            assert len(transformation_type) == n_transformations
            assert len(autoregressive_type) == n_transformations

        # initialize permutations:
        if permutations is None:
            _permutations = False
        elif isinstance(permutations, Iterable):
            assert len(permutations) == n_transformations
            _permutations = permutations
        elif isinstance(permutations, bool):
            if permutations:
                _permutations = min_var_permutations(d=num_params, n=n_transformations)
            else:
                _permutations = False
        self.permutations = _permutations

        # check type of architecture:
        if map_to_unitcube:
            assert transformation_type == 'spline'
            
        # check periodic parameters:
        if periodic_params is not None: 
            # if all parameters are not periodic then set to None:
            if np.all(np.logical_not(periodic_params)):
                periodic_params = None
        if periodic_params is not None: 
            assert transformation_type == 'spline'
            
        # check ranges for non-periodic parameters:
        if parameters_min is not None and parameters_max is not None:
            if transformation_type == 'spline':
                # the spline range better enclose all the samples:
                _temp_range_max = max(np.abs(np.amin(parameters_min)), np.abs(np.amax(parameters_max))).astype(type(range_max))
                if range_max < _temp_range_max:
                    if feedback > 0:
                        print('WARNING: range_max should be larger than the maximum range of the data and is beeing adjusted.')
                        print('    range_max:', range_max)
                        print('    max range:', _temp_range_max)
                    range_max = (_temp_range_max + 1).astype(np_prec)
                    if feedback > 0:
                        print('    new range_max:', range_max)
                    range_max = tf.cast(range_max, dtype=np_prec)
        # save ramge max:
        self.range_max = None
        if transformation_type == 'spline':
            self.range_max = range_max
        
        # initialize kernel initializer:    
        if kernel_initializer is None:
            kernel_initializer = tf.keras.initializers.VarianceScaling(
                scale=1. / n_transformations,
                mode='fan_avg',
                distribution='truncated_normal',
                seed=np.random.randint(np.iinfo(int_np_prec).max))

        # Build transformed distribution
        bijectors = []

        # handle first bijectors for periodic parameters:
        if periodic_params is not None:
            # apply shift, modulus and scaling bijector to periodic parameters:
            temp_bijectors = []
            for i in range(num_params):
                if periodic_params[i]:
                    temp_bijector_2 = []
                    # different strategies for splines and mafs:
                    if transformation_type == 'spline':
                        temp_bijector_2.append(tfb.Scale(1./range_max))
                    elif transformation_type == 'affine':
                        temp_bijector_2.append(tfb.Tanh())
                    # add bijector to list:
                    temp_bijectors.append(tfb.Chain(temp_bijector_2))
                else:
                    temp_bijectors.append(tfb.Identity())
            split = tfb.Split(num_params, axis=-1)
            bijectors.append(tfb.Chain([tfb.Invert(split), tfb.JointMap(temp_bijectors), split], name='PeriodicPreprocessing'))
        
        for i in range(n_transformations):

            # add permutations:
            if _permutations:
                if periodic_params is not None:
                    if i > 0:
                        bijectors.append(tfb.Permute(_permutations[i].astype(int_np_prec)))
                else:
                    bijectors.append(tfb.Permute(_permutations[i].astype(int_np_prec)))

            # add map to unit cube
            if map_to_unitcube:
                bijectors.append(tfb.Invert(tfb.NormalCDF()))

            # add main transformation
            _transformation_type = _transformation_types[i]
            _autoregressive_type = _autoregressive_types[i]

            if _transformation_type == 'affine':
                transf_params = 2
            elif _transformation_type == 'spline':
                # number of parameters:
                if periodic_params is not None:
                    transf_params = 3 * spline_knots
                else:
                    transf_params = 3 * spline_knots - 1
                # adjust for equispaced knots:
                if equispaced_x_knots:
                    transf_params -= spline_knots
                if equispaced_y_knots:
                    transf_params -= spline_knots
            else:
                raise ValueError

            ## first, get networks that parametrize transformation
            if _autoregressive_type == 'flex':
                nn = build_AR_model(
                    num_params,
                    transf_params,
                    hidden_units=hidden_units,
                    activation=activation,
                    kernel_initializer=kernel_initializer,
                    scale_with_dim=autoregressive_scale_with_dim,
                    identity_dims=autoregressive_identity_dims,
                    **utils.filter_kwargs(kwargs, Dense))
            elif _autoregressive_type == 'masked':
                nn = tfb.AutoregressiveNetwork(
                    params=transf_params,
                    event_shape=event_shape,
                    hidden_units=hidden_units,
                    activation=activation,
                    kernel_initializer=kernel_initializer,
                    **utils.filter_kwargs(kwargs, tfb.AutoregressiveNetwork))
            else:
                raise ValueError

            if _transformation_type == 'affine':
                transformation = tfb.MaskedAutoregressiveFlow(shift_and_log_scale_fn=nn)
            elif _transformation_type == 'spline':
                if map_to_unitcube:
                    transformation = SplineHelper(
                        shift_and_log_scale_fn=nn, spline_knots=spline_knots, range_min=0., range_max=1.,
                        **utils.filter_kwargs(kwargs, SplineHelper))
                else:
                    if periodic_params is not None:
                        transformation = CircularSplineHelper(
                            shift_and_log_scale_fn=nn, spline_knots=spline_knots, range_max=range_max,
                            equispaced_x_knots=equispaced_x_knots, equispaced_y_knots=equispaced_y_knots,
                            **utils.filter_kwargs(kwargs, CircularSplineHelper))
                    else:
                        transformation = SplineHelper(
                            shift_and_log_scale_fn=nn, spline_knots=spline_knots, range_max=range_max,
                            equispaced_x_knots=equispaced_x_knots, equispaced_y_knots=equispaced_y_knots,
                            **utils.filter_kwargs(kwargs, SplineHelper))
            bijectors.append(transformation)
            if map_to_unitcube:
                bijectors.append(tfb.NormalCDF())

            # add affine layer:
            if scale_roto_shift:
                bijectors.append(
                    ScaleRotoShift(
                        num_params,
                        name='affine_' + str(i) + '_' + str(np.random.randint(0, 100000)),
                        **utils.filter_kwargs(kwargs, ScaleRotoShift)))

        self.bijector = tfb.Chain(bijectors)

        if feedback > 1:
            print("    Building Autoregressive Flow")
            print("    - # parameters          :", num_params)
            print("    - periodic parameters   :", periodic_params)
            print("    - # transformations     :", n_transformations)
            print("    - hidden_units          :", hidden_units)
            print("    - transformation_type   :", transformation_type)
            print("    - autoregressive_type   :", autoregressive_type)
            print("    - permutations          :", permutations)
            print("    - scale_roto_shift      :", scale_roto_shift)
            print("    - activation            :", activation)

    def save(self, path):
        """
        Save a `AutoregressiveFlow` object.

        :param path: path of the directory where to save.
        :type path: str
        """
        checkpoint = tf.train.Checkpoint(bijector=self.bijector)
        checkpoint.write(path)
        with open(path + '_permutations.pickle', 'wb') as f:
            pickle.dump(self.permutations, f)
            
    @classmethod
    def load(cls, path, **kwargs):
        """
        Load a saved `AutoregressiveFlow` object. The number of parameters and all other keyword arguments (except for `permutations`) must be included as the MAF is first created with random weights and then these weights are restored.

        :type num_params: int
        :param path: path of the directory from which to load.
        :type path: str
        """
        permutations = pickle.load(open(path + '_permutations.pickle', 'rb'))
        maf = AutoregressiveFlow(
            num_params=len(permutations[0]),
            permutations=permutations,
            **kwargs)
        checkpoint = tf.train.Checkpoint(bijector=maf.bijector)
        checkpoint.read(path).expect_partial()
        return maf


class BijectorLayer(tf.keras.layers.Layer):
    """
    Custom Keras layer that applies a bijector transformation to the inputs.
    """

    def __init__(self, bijector, **kwargs):
        """
        Initializes the BijectorLayer.
        """
        super().__init__(**kwargs)
        self.bijector = bijector

    def call(self, inputs):
        """
        Applies the forward transformation of the bijector to the inputs.
        """
        return self.bijector.forward(inputs)





"""
This file contains analytic fixed bijectors that are not trainable.
These are used to pre-Gaussianize the prior distribution and to perform
other operations on the distribution.

These bijectors are meant to be fixed and not trainable.
"""

###############################################################################
# initial imports and set-up:

import numpy as np

# tensorflow imports:
import tensorflow as tf
import tensorflow_probability as tfp
tfb = tfp.bijectors
tfd = tfp.distributions

# default precision
np_prec = np.float32

###############################################################################
# Definitions of analytic bijectors for Gaussianizing the prior:

def uniform_prior(a, b, prec=np_prec):
    """
    Returns bijector that Gaussianize the 1D uniform distribution in [a, b].

    Parameters:
    a (float): The lower bound of the uniform distribution.
    b (float): The upper bound of the uniform distribution.
    prec (function, optional): The precision function used for numerical stability. Defaults to np_prec.

    Returns:
    tfb.Chain: The bijector Gaussianizing the uniform distribution.
    """
    return tfb.Chain([tfb.Shift(prec((a+b)/2)), tfb.Scale(prec(b-a)), tfb.Shift(-0.5), tfb.NormalCDF()])


def normal(mean, sigma, prec=np_prec):
    """
    Returns bijector that Gaussianize the 1D normal distribution with mean and variance sigma.
    This ammounts to a shift and rescaling since the normal distribution is already Gaussian.

    Parameters:
    mean (float): The mean of the normal distribution.
    sigma (float): The standard deviation of the normal distribution.
    prec (float, optional): The precision parameter. Defaults to np_prec.

    Returns:
    tfb.Chain: The bijector for the normal distribution.
    """
    return tfb.Chain([tfb.Shift(prec(mean)), tfb.Scale(prec(sigma))])


def multivariate_normal(mean, covariance, prec=np_prec):
    """
    Returns bijector to Gaussianize the ND normal distribution with mean mu and given covariance.
    This ammounts to a shift and rescaling since the normal distribution is already Gaussian.

    Parameters:
    mean (array-like): Mean of the normal distribution.
    covariance (array-like): Covariance matrix of the normal distribution.
    prec (dtype, optional): Precision type for numerical computations. Defaults to np_prec.

    Returns:
    tfb.Bijector: Bijector for the multivariate normal distribution.
    """
    return tfd.MultivariateNormalTriL(mean=mean.astype(prec), scale_tril=tf.linalg.cholesky(covariance.astype(prec))).bijector


###############################################################################
# helper function to generate analytic prior bijectors


def prior_bijector_helper(prior_dict_list=None, name=None, loc=None, cov=None, **kwargs):
    """
    Example usage

    # prior_dict_list should contain a mode keyword. If mode is 'uniform', then
    # the prior is uniform on the interval [lower, upper]. If mode is 'gaussian',
    # then the prior is gaussian with mean and scale.
    
    # uniform on x
    a = -1
    b = 3

    # gaussian on y
    mu = 0.5
    sig = 3.

    prior = prior_bijector_helper([{'lower':a, 'upper':b}, {'mean':mu, 'scale':sig}])
    diff = FlowCallback(chain, trainable_bijector=prior, Y2X_is_identity=True)

    """
    if prior_dict_list is not None:  # Mix of uniform and gaussian one-dimensional priors

        # Build one-dimensional bijectors
        n = len(prior_dict_list)
        temp_bijectors = []
        for i in range(n):
            
            if prior_dict_list[i] is not None:
                if 'mode' in prior_dict_list[i].keys():
                    if prior_dict_list[i]['mode'] == 'uniform':
                        temp_bijectors.append(uniform_prior(prior_dict_list[i]['lower'], prior_dict_list[i]['upper']))
                    elif prior_dict_list[i]['mode'] == 'gaussian':
                        temp_bijectors.append(normal(prior_dict_list[i]['mean'], prior_dict_list[i]['scale']))
                    else:
                        raise ValueError
            else:
                temp_bijectors.append(tfb.Identity())       
            
        # Need Split() to split/merge inputs
        split = tfb.Split(n, axis=-1)

        # Chain all
        return tfb.Chain([tfb.Invert(split), tfb.JointMap(temp_bijectors), split], name=name)

    elif loc is not None:  # Multivariate Gaussian prior
        assert cov is not None
        return multivariate_normal(loc, cov)
    else:
        raise ValueError

###############################################################################
# definition of the fixed modulus bijector


class Mod1D(tfb.Bijector):
    """
    A bijector that performs modulus operation on a 1D input.

    This bijector maps an input `x` to `x - floor((x - minval) / delta) * delta`,
    where `minval` is the lower bound of the modulus and `delta` is the difference
    between the upper and lower bounds.

    Args:
        minval (float): The lower bound of the modulus. Default is 0.0.
        maxval (float): The upper bound of the modulus. Default is 1.0.
        validate_args (bool): Whether to validate input arguments. Default is False.
        name (str): The name of the bijector. Default is 'mod'.
        dtype (tf.DType): The data type of the input. Default is np_prec.

    """

    def __init__(
            self,
            minval=0.0,
            maxval=1.0,
            validate_args=False,
            name='mod',
            dtype=np_prec):
        """
        Initializes the Mod1D bijector.

        Args:
            minval (float): The lower bound of the modulus.
            maxval (float): The upper bound of the modulus.
            validate_args (bool): Whether to validate input arguments.
            name (str): The name of the bijector.
            dtype (tf.DType): The data type of the input.
        """

        parameters = dict(locals())
        
        self.delta = maxval - minval
        self.minval = minval
        self.maxval = maxval

        with tf.name_scope(name) as name:

            super(Mod1D, self).__init__(
                forward_min_event_ndims=0,
                is_constant_jacobian=True,
                validate_args=validate_args,
                parameters=parameters,
                dtype=dtype,
                name=name)

    @classmethod
    def _is_increasing(cls):
        return True

    def _forward(self, x):
        return x - tf.math.floor((x-self.minval)/self.delta)*self.delta

    def _inverse(self, y):
        return y - tf.math.floor((y-self.minval)/self.delta)*self.delta
    
    def _forward_log_det_jacobian(self, x):
        return tf.zeros(tf.shape(x))

    def _inverse_log_det_jacobian(self, y):
        return tf.zeros(tf.shape(y))
    


"""
This module contains the functions and utilities to compute non-Gaussian
Monte Carlo tension estimators.

The submodule `param_diff` contains the functions and utilities to compute the distribution
of parameter differences from the parameter posterior of two experiments.

The submodule `kde` contains the functions to compute the statistical significance
of a difference in parameters with KDE methods.

This submodule `flow` contains the functions and utilities to compute the statistical significance
of a difference in parameters with normalizing flow methods.

For more details on the method implemented see
`arxiv 1806.04649 <https://arxiv.org/pdf/1806.04649.pdf>`_
and `arxiv 2105.03324 <https://arxiv.org/pdf/2105.03324.pdf>`_.
"""

# parameter difference module import:
from .param_diff import parameter_diff_chain, parameter_diff_weighted_samples

# kde module import:
from .kde import kde_parameter_shift_1D_fft, kde_parameter_shift_2D_fft, kde_parameter_shift

# flow module import:
from .flow import estimate_shift, estimate_shift_from_samples, flow_parameter_shift



"""
This file contains a set of utilities operating on flows.
"""

###############################################################################
# initial imports and set-up:

import numpy as np

from getdist import MCSamples

###############################################################################
# get samples at each intermediate space:

def get_samples_bijectors(flow, feedback=False, extra_samples=None):
    
    # initialize the flow:
    _flow = flow
    # initialize lists to store samples:
    training_samples_spaces = []
    validation_samples_spaces = []
    if extra_samples is not None:
        extra_samples_spaces = []
    # initialize MCSamples with original space samples:
    training_samples_spaces.append(MCSamples(samples=_flow.chain_samples[_flow.training_idx, :],
                                            weights=_flow.chain_weights[_flow.training_idx],
                                            name_tag='original_space',
                                            ))
    validation_samples_spaces.append(MCSamples(samples=_flow.chain_samples[_flow.test_idx, :],
                                            weights=_flow.chain_weights[_flow.test_idx],
                                            name_tag='original_space',
                                            ))
    if extra_samples is not None:
        extra_samples_spaces.append(extra_samples)
    # append MCSamples with training samples:
    training_samples_spaces.append(MCSamples(samples=_flow.training_samples,
                                            weights=_flow.training_weights,
                                            name_tag='training_space',
                                            ))
    validation_samples_spaces.append(MCSamples(samples=_flow.test_samples,
                                            weights=_flow.test_weights,
                                            name_tag='validation_space',
                                            ))
    if extra_samples is not None:
        extra_samples_spaces.append(flow.fixed_bijector.inverse(extra_samples).numpy())
    # loop over bijectors:
    _temp_train_samples = _flow.training_samples
    _temp_test_samples = _flow.test_samples
    if extra_samples is not None:
        _temp_extra_samples = flow.fixed_bijector.inverse(extra_samples).numpy()
    for ind, bijector in enumerate(_flow.trainable_bijector._bijectors):
        # feedback:
        if feedback:
            print(ind, '- bijector name: ', bijector.name)
        # process samples trough the bijector:
        _temp_train_samples = bijector.inverse(_temp_train_samples)
        _temp_test_samples = bijector.inverse(_temp_test_samples)
        if extra_samples is not None:
            _temp_extra_samples = bijector.inverse(_temp_extra_samples)
            extra_samples_spaces.append(_temp_extra_samples.numpy())
        # get the training samples:
        training_samples_spaces.append(MCSamples(samples=_temp_train_samples.numpy(),
                                                weights=_flow.training_weights,
                                                name_tag=str(ind)+'_after_'+bijector.name,
                                                ))
        # get the validation samples:
        validation_samples_spaces.append(MCSamples(samples=_temp_test_samples.numpy(),
                                                weights=_flow.test_weights,
                                                name_tag=str(ind)+'_after_'+bijector.name,
                                                ))
    # return the samples:
    if extra_samples is not None:
        return training_samples_spaces, validation_samples_spaces, extra_samples_spaces
    else:
        return training_samples_spaces, validation_samples_spaces

###############################################################################
# flow KL divergence:

def KL_divergence(flow_1, flow_2, num_samples=1000, num_batches=100):
    """
    Calculates the Kullback-Leibler (KL) divergence between two flows:
    
    D_KL(flow_1 || flow_2) = E_{x ~ flow_1} [log(flow_1(x)) - log(flow_2(x))]

    Parameters:
    flow_1 (Flow): The flow for the sampling distribution.
    flow_2 (Flow): The second flow distribution.
    num_samples (int): The number of samples to draw from the flows, per batch. Default is 1000.
    num_batches (int): The number of batches to run. Default is 100.

    Returns:
    mean (float): The mean KL divergence.
    std (float): The standard deviation of the KL divergence.
    """
    _log_prob_diff = []
    for _ in range(num_batches):
        # sample from the first flow:
        _temp_samples = flow_1.sample(num_samples)
        # calculate the log probability difference:
        _temp_diff = flow_1.log_probability(_temp_samples) - flow_2.log_probability(_temp_samples)
        # filter out not finite values:
        _temp_diff = _temp_diff[np.isfinite(_temp_diff)]
        # average and append:        
        _log_prob_diff.append(np.mean(_temp_diff))
    # convert to numpy array:
    _log_prob_diff = np.array(_log_prob_diff)
    # filter out not finite values:
    _log_prob_diff = _log_prob_diff[np.isfinite(_log_prob_diff)]
    #
    return np.mean(_log_prob_diff), np.std(_log_prob_diff)


"""
This module contains functions to estimate the probability of a parameter shift given a 
synthetic probability model of the parameter difference distribution.

For further details we refer to `arxiv 2105.03324 <https://arxiv.org/pdf/2105.03324.pdf>`_.
"""

###############################################################################
# initial imports and set-up:

import numpy as np

from .. import utilities as utils

###############################################################################
# function to estimate the flow probability of zero shift:


def estimate_shift(flow, prior_flow=None, tol=0.05, max_iter=1000, step=100000):
    """
    Compute the normalizing flow estimate of the probability of a parameter shift 
    given the input parameter difference chain. 
    This is done with a Monte Carlo estimate by comparing the probability density 
    at the zero-shift point to that at samples drawn from the normalizing flow 
    approximation of the distribution.

    :param flow: the input flow for a parameter difference distribution.
    :param prior_flow: the input flow for the prior distribution, defaults to None.
    :param tol: absolute tolerance on the shift significance, defaults to 0.05.
    :type tol: float, optional
    :param max_iter: maximum number of sampling steps, defaults to 1000.
    :type max_iter: int, optional
    :param step: number of samples per step, defaults to 100000.
    :type step: int, optional
    :return: probability value and error estimate.
    """
    err = np.inf
    counter = max_iter

    # define threshold for tension calculation:
    _thres = flow.log_probability(flow.cast(np.zeros(flow.num_params)))
    if prior_flow is not None:
        _thres = _thres - prior_flow.log_probability(prior_flow.cast(np.zeros(prior_flow.num_params)))

    _num_filtered = 0
    _num_samples = 0
    while err > tol and counter >= 0:
        counter -= 1
        # sample from the flow:
        _s = flow.sample(step)
        # compute probability values:
        _s_prob = flow.log_probability(_s)
        if prior_flow is not None:
            _s_prob = _s_prob - prior_flow.log_probability(prior_flow.cast(_s))
        # test:
        _t = np.array(_s_prob > _thres)
        # update counters:
        _num_filtered += np.sum(_t)
        _num_samples += step
        _P = float(_num_filtered)/float(_num_samples)
        _low, _upper = utils.clopper_pearson_binomial_trial(float(_num_filtered),
                                                            float(_num_samples),
                                                            alpha=0.32)

        err = np.abs(utils.from_confidence_to_sigma(_upper)-utils.from_confidence_to_sigma(_low))

    return _P, _low, _upper


def estimate_shift_from_samples(flow, prior_flow=None):
    """
    Compute the normalizing flow estimate of the probability of a parameter shift 
    given the input parameter difference chain. 
    This is done with a Monte Carlo estimate by comparing the probability density 
    at the zero-shift point to that at samples drawn from the normalizing flow 
    approximation of the distribution.
    This function does not use the flow to sample but rather uses the input samples.
    As a downside there is no control on the end accuracy of the estimate.
    
    :param flow: the input flow for a parameter difference distribution.
    :param prior_flow: the input flow for the prior distribution, defaults to None.
    :param tol: absolute tolerance on the shift significance, defaults to 0.05.
    :type tol: float, optional
    :param max_iter: maximum number of sampling steps, defaults to 1000.
    :type max_iter: int, optional
    :param step: number of samples per step, defaults to 100000.
    :type step: int, optional
    :return: probability value and error estimate.
    """

    # define threshold for tension calculation:
    _thres = flow.log_probability(flow.cast(np.zeros(flow.num_params)))
    if prior_flow is not None:
        _thres = _thres - prior_flow.log_probability(prior_flow.cast(np.zeros(prior_flow.num_params)))

    # calculate probability on the samples:
    _s_prob = flow.log_probability(flow.cast(flow.chain_samples))
    if prior_flow is not None:
        _s_prob = _s_prob - prior_flow.log_probability(prior_flow.cast(flow.chain_samples))

    # calculate probability of shift:
    _t = np.array(_s_prob > _thres)
    _num_filtered = np.sum(_t)
    _num_samples = len(_t)
    _P = float(_num_filtered)/float(_num_samples)
    _low, _upper = utils.clopper_pearson_binomial_trial(float(_num_filtered),
                                                        float(_num_samples),
                                                        alpha=0.32)

    return _P, _low, _upper

###############################################################################
# helper function to compute tension with default MAF:


def flow_parameter_shift(diff_chain, cache_dir=None, root_name='sprob', tol=0.05, max_iter=1000, step=100000, **kwargs):
    """
    Wrapper function to compute a normalizing flow estimate of the probability of a parameter shift given the input 
    parameter difference chain. 
    The function accepts as kwargs all the ones that are relevant for the function flow_from_chain.

    :param diff_chain: input parameter difference chain.
    :type diff_chain: :class:`~getdist.mcsamples.MCSamples`
    :param cache_dir: name of the directory to save training cache files. If none (default) does not cache.
    :param root_name: root name for the cache files.
    :param tol: absolute tolerance on the shift significance, defaults to 0.05.
    :type tol: float, optional
    :param max_iter: maximum number of sampling steps, defaults to 1000.
    :type max_iter: int, optional
    :param step: number of samples per step, defaults to 100000.
    :type step: int, optional
    :return: probability value and error estimate, then the parameter difference flow
    """
    # since this function uses the synthetic_probability module, we need to import it here:
    from ..synthetic_probability import synthetic_probability as sp
    # initialize and train parameter difference flow:
    diff_flow = sp.flow_from_chain(diff_chain, cache_dir=cache_dir, root_name=root_name, **kwargs)
    # Compute tension:
    result = estimate_shift(diff_flow, tol=tol, max_iter=max_iter, step=step)
    #
    return result, diff_flow



"""
This file contains the loss functions for the normalizing flow training.

Since we are combining different loss functions we have different options.
"""

###############################################################################
# initial imports and set-up:

import numpy as np

# tensorflow imports:
import tensorflow as tf
from tensorflow.python.keras.utils import tf_utils
from tensorflow.python.keras import backend
from tensorflow.python.keras.utils import losses_utils

###############################################################################
# standard normalizing flow loss function:


class standard_loss(tf.keras.losses.Loss):

    def __init__(self):
        """
        Standard density loss function for the normalizing flow.
        """
        # initialize:
        super(standard_loss, self).__init__()
        #
        return None

    def compute_loss_components(self, y_true, y_pred, sample_weight):
        """
        Compute different components of the loss function
        """
        return -y_pred

    def call(self, y_true, y_pred):
        """
        Standard normalizing flow loss function is KL divergence of two abstract
        distributions.
        """
        return -y_pred
    
    def print_feedback(self, padding=''):
        """
        Print feedback to screen
        """
        print(padding+'using standard loss function')

    def reset(self):
        """
        Reset loss functions hyper parameters
        """
        pass


###############################################################################
# density and evidence loss with constant weights:


class constant_weight_loss(tf.keras.losses.Loss):

    def __init__(self, alpha=1.0, beta=0.0):
        """
        Initialize loss function
        """
        # initialize:
        super(constant_weight_loss, self).__init__()
        # set parameters:
        self.alpha = alpha
        self.beta = beta
        #
        return None

    def compute_loss_components(self, y_true, y_pred, sample_weight):
        """
        Compute different components of the loss function
        """
        # compute difference between true and predicted posterior values:
        diffs = (y_true - y_pred)
        # sum weights:
        tot_weights = tf.reduce_sum(sample_weight)
        # compute overall offset:
        mean_diff = tf.reduce_sum(diffs*sample_weight) / tot_weights
        # compute its variance:
        var_diff = (diffs - mean_diff)**2
        # compute density loss function:
        loss_orig = -(y_pred + self.beta)
        #
        return loss_orig, var_diff

    def compute_loss(self, y_true, y_pred, sample_weight):
        """
        Combine density and evidence-error loss
        """
        # get components:
        loss_1, loss_2 = self.compute_loss_components(y_true, y_pred, sample_weight)
        #
        return +self.alpha*loss_1 + (1. - self.alpha)*loss_2

    def __call__(self, y_true, y_pred, sample_weight=None):
        """
        This function overrides the standard tensorflow one to pass along weights.
        """
        graph_ctx = tf_utils.graph_context_for_symbolic_tensors(
            y_true, y_pred, sample_weight
            )
        with backend.name_scope(self._name_scope), graph_ctx:
            if tf.executing_eagerly():
                call_fn = self.compute_loss
            else:
                call_fn = tf.__internal__.autograph.tf_convert(self.compute_loss, tf.__internal__.autograph.control_status_ctx())
            losses = call_fn(y_true, y_pred, sample_weight)
            return losses_utils.compute_weighted_loss(
                losses, sample_weight, reduction=self._get_reduction()
                )

    def print_feedback(self, padding=''):
        """
        Print feedback to screen
        """
        print(padding+'using combined density and evidence-error loss function')
        print(padding+'weight of density loss: %.3g, weight of evidence-error-loss: %.3g' % (self.alpha, 1.-self.alpha))

    def reset(self):
        """
        Reset loss functions hyper parameters
        """
        pass

###############################################################################
# density and evidence loss with variable weights:


class variable_weight_loss(tf.keras.losses.Loss):

    def __init__(self, lambda_1=1.0, lambda_2=0.0, beta=0.0):
        """
        Initialize loss function
        """
        # initialize:
        super(variable_weight_loss, self).__init__()
        # set parameters:
        self.lambda_1 = tf.Variable(lambda_1, trainable=False, name='loss_lambda_1', dtype=type(lambda_1))
        self.lambda_2 = tf.Variable(lambda_2, trainable=False, name='loss_lambda_2', dtype=type(lambda_2))
        self.beta = tf.Variable(beta, trainable=False, name='loss_beta', dtype=type(beta))
        # save initial parameters:
        self.initial_lambda_1 = lambda_1
        self.initial_lambda_2 = lambda_2
        self.initial_beta = beta
        #
        return None

    def update_lambda_values_on_epoch_begin(self, epoch, **kwargs):
        """
        Update values of lambda at epoch start. Takes in every kwargs to not
        crowd the interface...
        """
        # base class is empty...
        # use the following sintax:
        # tf.keras.backend.set_value(self.lambda_1, tf.constant(0.5) * epoch)
        raise NotImplementedError

    def compute_loss_components(self, y_true, y_pred, sample_weight, lambda_1=None, lambda_2=None):
        """
        Compute different components of the loss function
        """
        # compute difference between true and predicted posterior values:
        diffs = (y_true - y_pred)
        # sum weights:
        tot_weights = tf.reduce_sum(sample_weight)
        # compute overall offset:
        mean_diff = tf.reduce_sum(diffs*sample_weight) / tot_weights
        # compute its variance:
        var_diff = (diffs - mean_diff)**2
        # compute density loss function:
        loss_orig = -(y_pred + self.beta)
        # get weights if not passed:
        if lambda_1 is None:
            lambda_1 = tf.keras.backend.get_value(self.lambda_1)
        if lambda_2 is None:
            lambda_2 = tf.keras.backend.get_value(self.lambda_2)
        #
        return loss_orig, var_diff, lambda_1, lambda_2

    def compute_loss(self, y_true, y_pred, sample_weight):
        """
        Combine density and evidence-error loss
        """
        # get components:
        loss_1, loss_2, lambda_1,lambda_2 = self.compute_loss_components(y_true, y_pred, sample_weight, self.lambda_1, self.lambda_2)
        #
        return lambda_1*loss_1 + lambda_2*loss_2

    def __call__(self, y_true, y_pred, sample_weight=None):
        """
        This function overrides the standard tensorflow one to pass along weights.
        """
        graph_ctx = tf_utils.graph_context_for_symbolic_tensors(
            y_true, y_pred, sample_weight
            )
        with backend.name_scope(self._name_scope), graph_ctx:
            if tf.executing_eagerly():
                call_fn = self.compute_loss
            else:
                call_fn = tf.__internal__.autograph.tf_convert(self.compute_loss, tf.__internal__.autograph.control_status_ctx())
            losses = call_fn(y_true, y_pred, sample_weight)
            return losses_utils.compute_weighted_loss(
                losses, sample_weight, reduction=self._get_reduction()
                )
        
    def print_feedback(self, padding=''):
        """
        Print feedback to screen
        """
        raise NotImplementedError

    def reset(self):
        """
        Reset loss functions hyper parameters
        """
        self.__init__(lambda_1=self.initial_lambda_1, lambda_2=self.initial_lambda_2, beta=self.initial_beta)


class random_weight_loss(variable_weight_loss):
    """
    Random weighting of the two loss functions.
    """

    def __init__(self, initial_random_epoch=0, lambda_1=1.0, beta=0.0, **kwargs):
        """
        Initialize loss function
        """
        # initialize:
        super(random_weight_loss, self).__init__(lambda_1, 1.-lambda_1, beta)
        # set parameters:
        self.initial_random_epoch = initial_random_epoch
        #
        return None

    def update_lambda_values_on_epoch_begin(self, epoch, **kwargs):
        """
        Update values of lambda at epoch start. Takes in every kwargs to not
        crowd the interface...
        """
        if epoch > self.initial_random_epoch:
            _temp_rand = np.random.randint(2)
            tf.keras.backend.set_value(self.lambda_1, _temp_rand)
            tf.keras.backend.set_value(self.lambda_2, 1.-_temp_rand)
        #        
        return None

    def print_feedback(self, padding=''):
        """
        Print feedback to screen
        """
        print(padding+'using randomized loss function')


class annealed_weight_loss(variable_weight_loss):
    """
    Slowly go from density to evidence-error loss.
    """

    def __init__(self, anneal_epoch=125, lambda_1=1.0, beta=0.0, roll_off_nepoch=10, **kwargs):
        """
        Initialize loss function
        """
        # initialize:
        super(annealed_weight_loss, self).__init__(lambda_1, 1.-lambda_1, beta)
        # set parameters:
        self.anneal_epoch = anneal_epoch
        self.roll_off_nepoch = roll_off_nepoch
        #
        return None

    def update_lambda_values_on_epoch_begin(self, epoch, **kwargs):
        """
        Update values of lambda at epoch start. Takes in every kwargs to not
        crowd the interface...
        """
        if epoch > self.anneal_epoch:
            _lambda_1 = tf.keras.backend.get_value(self.lambda_1)
            _lambda_1 *= np.exp(-1.*(epoch - self.anneal_epoch)/self.roll_off_nepoch)
            tf.keras.backend.set_value(self.lambda_1, _lambda_1)
            tf.keras.backend.set_value(self.lambda_2, 1.-_lambda_1)
        #
        return None

    def print_feedback(self, padding=''):
        """
        Print feedback to screen
        """
        print(padding+'using annealed loss function')


class SoftAdapt_weight_loss(variable_weight_loss):
    """
    Implement SoftAdapt as in arXiv:1912.12355, with optional smoothing
    """

    def __init__(self, tau=1.0, beta=0.0, smoothing=True, smoothing_tau=20, quantity_1='val_rho_loss', quantity_2='val_ee_loss', **kwargs):
        """
        Initialize loss function
        """
        # initialize:
        super(SoftAdapt_weight_loss, self).__init__()
        # set parameters:
        self.tau = tau
        self.beta = beta
        self.smoothing = smoothing
        self.smoothing_alpha = 1. / smoothing_tau
        # quantrities to be monitored:
        self.quantity_1 = quantity_1 
        self.quantity_2 = quantity_2 
        self.rate_1_buffer = 0.0
        self.rate_2_buffer = 0.0
        self.lambda_1_buffer = 1.0
        self.lambda_2_buffer = 0.0
        #
        return None

    def update_lambda_values_on_epoch_begin(self, epoch, **kwargs):
        """
        Update values of lambda at epoch start. Takes in every kwargs to not
        crowd the interface...
        """
        # get logs:
        logs = kwargs.get('logs')
        quantity_1 = logs[self.quantity_1]
        quantity_2 = logs[self.quantity_2]
        # get the two rates:
        if len(quantity_1) < 2:
            rate_1 = 0.0
        else:
            rate_1 = quantity_1[-1] - quantity_1[-2]
        if len(quantity_2) < 2:
            rate_2 = 0.0
        else:
            rate_2 = quantity_2[-1] - quantity_2[-2]
        # smooth the two rates:
        if self.smoothing:
            rate_1 = self.smoothing_alpha * rate_1 + (1. - self.smoothing_alpha) * self.rate_1_buffer
            rate_2 = self.smoothing_alpha * rate_2 + (1. - self.smoothing_alpha) * self.rate_2_buffer
            self.rate_1_buffer = rate_1
            self.rate_2_buffer = rate_2
        # protect for initial phase:
        if rate_1 == 0.0:
            _lambda_1 = 1.0
        else:
            lambda_1 = np.exp(self.tau * rate_1)
            lambda_2 = np.exp(self.tau * rate_2)
            _tot = lambda_1 + lambda_2
            if self.smoothing:
                _lambda_1 = self.smoothing_alpha * lambda_1 / _tot + (1. - self.smoothing_alpha) * self.lambda_1_buffer
                _lambda_2 = self.smoothing_alpha * lambda_2 / _tot + (1. - self.smoothing_alpha) * self.lambda_2_buffer
                _lambda_1 = _lambda_1 / (_lambda_1 + _lambda_2)
                _lambda_2 = _lambda_2 / (_lambda_1 + _lambda_2)
                self.lambda_1_buffer = _lambda_1
                self.lambda_2_buffer = _lambda_2
            else:
                _lambda_1 = lambda_1 / _tot
        # set second by enforcing sum to one:
        tf.keras.backend.set_value(self.lambda_1, _lambda_1)
        tf.keras.backend.set_value(self.lambda_2, 1.-_lambda_1)
        #
        return None

    def print_feedback(self, padding=''):
        """
        Print feedback to screen
        """
        print(padding+'using SoftAdapt loss function')
        if self.smoothing:
           print(padding+' with smoothing')
           print(padding+' with smoothing_tau = ', 1./self.smoothing_alpha)
        print(padding+' with tau = ', self.tau)


class SharpStep(variable_weight_loss):
    """
    Implement sharp stepping between two values
    """

    def __init__(self, step_epoch=50, value_1=1.0, value_2=0.1, beta=0., **kwargs):
        """
        Initialize loss function
        """
        # initialize:
        super(SharpStep, self).__init__()
        # set parameters:
        self.step_epoch = step_epoch
        self.value_1 = value_1
        self.value_2 = value_2
        # initialize:
        self.beta = beta
        #
        return None

    def update_lambda_values_on_epoch_begin(self, epoch, **kwargs):
        """
        Update values of lambda at epoch start. Takes in every kwargs to not
        crowd the interface...
        """
        if epoch < self.step_epoch:
            lambda_1 = self.value_1
        else:
            lambda_1 = self.value_2
        # set second by enforcing sum to one:
        lambda_2 = 1. - lambda_1
        #
        tf.keras.backend.set_value(self.lambda_1, lambda_1)
        tf.keras.backend.set_value(self.lambda_2, lambda_2)
        #
        return None

    def print_feedback(self, padding=''):
        """
        Print feedback to screen
        """
        print(padding+'using sharp step loss function')
